// Module for DirectInput
#include "app.h"

static IDirectInput* pDI = NULL;					// DirectInput interface
static HWND hWnd = NULL;							// The window handle to use for DirectInput

#define MAX_JOYSTICKS (64)
static IDirectInputDevice2* JoyList[MAX_JOYSTICKS];
static int* JoyPrevAxes = NULL;
static IDirectInputDevice* Keyboard = NULL;

// Number of joysticks connected to this machine
static int nJoystickCount=0;

static DIJOYSTATE *pCalcState = NULL;
#define DINPUT_BUTTON_COUNT (sizeof(pCalcState->rgbButtons) / sizeof(pCalcState->rgbButtons[0]))

// This function is called back with each joystick connected to the system
static BOOL CALLBACK DIEnumDevicesProc(LPCDIDEVICEINSTANCE pDDI, LPVOID /*pvRef*/)
{
	int nRet = 0;
	IDirectInputDevice* pDev = NULL;
	IDirectInputDevice2* pDev2 = NULL;

	if (nJoystickCount >= MAX_JOYSTICKS) {			// Already have the maximum number of joysticks
		return DIENUM_STOP;
	}
	if (pDDI == NULL) {
		return DIENUM_CONTINUE;
	}

	// Create the DirectInput1 interface
	nRet = pDI->CreateDevice(pDDI->guidInstance, &pDev, NULL);
	if (nRet < 0 || pDev == NULL) {
		return DIENUM_CONTINUE;
	}

	// Get the DirectInputDevice2 interface so we can use the Poll method
	nRet = pDev->QueryInterface(IID_IDirectInputDevice2, (void**)&pDev2);
	RELEASE(pDev)
	if (nRet < 0 || pDev2 == NULL) {
		return DIENUM_CONTINUE;
	}

	// We now have a IDirectInputDevice2 interface to store
	JoyList[nJoystickCount++] = pDev2;

	return DIENUM_CONTINUE;
}

// Sets up one Joystick (for example the range of the joystick's axes)
static int SetUpJoystick(int i)
{
	int nRet = 0;
	IDirectInputDevice2* pJoy = NULL;
	DIPROPRANGE diprg;
	
	pJoy = JoyList[i];
	if (pJoy == NULL) {
		return 1;
	}

	nRet = pJoy->SetDataFormat(&c_dfDIJoystick);
	if (nRet < 0) {
		return 1;
	}
	
	pJoy->SetCooperativeLevel(hWnd, DISCL_EXCLUSIVE | DISCL_BACKGROUND);

	// Make range is -0x10000 to 0x10000
	memset(&diprg, 0, sizeof(diprg));
	diprg.diph.dwSize = sizeof(diprg);
	diprg.diph.dwHeaderSize = sizeof(diprg.diph);
	diprg.diph.dwHow = DIPH_BYOFFSET;
	diprg.lMin = -0x10000;
	diprg.lMax = 0x10000;	
	
	diprg.diph.dwObj = DIJOFS_X;
	pJoy->SetProperty(DIPROP_RANGE, &diprg.diph);
	diprg.diph.dwObj = DIJOFS_Y;	
	pJoy->SetProperty(DIPROP_RANGE, &diprg.diph);
	diprg.diph.dwObj = DIJOFS_Z;
	pJoy->SetProperty(DIPROP_RANGE, &diprg.diph);
	diprg.diph.dwObj = DIJOFS_RX;
	pJoy->SetProperty(DIPROP_RANGE, &diprg.diph);
	diprg.diph.dwObj = DIJOFS_RY;
	pJoy->SetProperty(DIPROP_RANGE, &diprg.diph);
	diprg.diph.dwObj = DIJOFS_RZ;
	pJoy->SetProperty(DIPROP_RANGE, &diprg.diph);
	diprg.diph.dwObj = DIJOFS_SLIDER(0);
	pJoy->SetProperty(DIPROP_RANGE, &diprg.diph);
	diprg.diph.dwObj = DIJOFS_SLIDER(1);
	pJoy->SetProperty(DIPROP_RANGE, &diprg.diph);

	pJoy->Acquire();

	return 0;
}

static int DirKeyboardInit()
{
	// Get a DI interface to the keyboard
	int nRet = 0;
	Keyboard = NULL;
	nRet = pDI->CreateDevice(GUID_SysKeyboard, &Keyboard, NULL);
	if (nRet < 0 || Keyboard == NULL) {
		return 1;
	}
	
	Keyboard->SetDataFormat(&c_dfDIKeyboard);
	Keyboard->SetCooperativeLevel(hWnd, DISCL_NONEXCLUSIVE | DISCL_BACKGROUND);
	Keyboard->Acquire();
	
	return 0;
}

int DinpInit(HINSTANCE hInst,HWND hPassWnd)
{
	int nRet = 0;
	hWnd = hPassWnd;
	int nSize;

	// Make sure all Interfaces are NULL
	nJoystickCount = 0;
	
	free(JoyPrevAxes);

	memset(&JoyList, 0, sizeof(JoyList));

	nSize = MAX_JOYSTICKS * 8 * sizeof(int);
	if ((JoyPrevAxes = (int*)malloc(nSize)) == NULL) {
		DinpExit();
		return 1;
	}
	memset(JoyPrevAxes, 0, nSize);

	nRet = DirectInputCreate(hInst, DIRECTINPUT_VERSION, &pDI, NULL);
	if (nRet < 0) {
		return 1;
	}

	// enumerate the joysticks connected to the system
	pDI->EnumDevices(DIDEVTYPE_JOYSTICK, DIEnumDevicesProc, NULL, DIEDFL_ATTACHEDONLY);


	// Set up the DI2 devices
	for (int i = 0; i < nJoystickCount; i++) {
		SetUpJoystick(i);
	}

	// Set up the keyboard
	DirKeyboardInit();

	return 0;
}

int DinpExit()
{
	// Release all the Effect/DI2 interfaces
	for (int i = 0; i < MAX_JOYSTICKS; i++) {
		RELEASE(JoyList[i])
	}
	nJoystickCount = 0;

	// Release the keyboard interface
	RELEASE(Keyboard)

	// Release the DirectInput interface
	RELEASE(pDI)
	
	free(JoyPrevAxes);
	JoyPrevAxes = NULL;

	return 0;
}

static unsigned char bJoystickRead[MAX_JOYSTICKS];
static DIJOYSTATE JoyState[MAX_JOYSTICKS];
static int bKeyboardRead = 0;						// A flag to say if the Keyboard was read this frame
static unsigned char KeyboardState[256];

int DinpStart()										// Call before checking for Input in a frame
{
	// No joysticks have been Read for this frame
	memset(&bJoystickRead, 0, nJoystickCount * sizeof(bJoystickRead[0]));
	
	// Keyboard not Read this frame
	bKeyboardRead = 0;
	
	return 0;
}

// Read one of the joysticks
static inline int ReadJoystick(int i)
{
	int nRet = 0;
	IDirectInputDevice2 *pJoy = NULL;
	
	if (bJoystickRead[i] == 1) {					// already read this frame - ready to go
		return 0;
	}
	if (bJoystickRead[i] == 2) {					// error reading joystick
		return 1;
	}

	pJoy = JoyList[i];

	bJoystickRead[i] = 2;							// Mark joystick as trying to read

	// Poll + read joystick, reacquiring if neccesary
	for (int j = 0; j < 2; j++) {
		pJoy->Poll();

		nRet = pJoy->GetDeviceState(sizeof(JoyState[i]), &JoyState[i]);
		if (nRet != DIERR_INPUTLOST) {				
			break;
		}
		pJoy->Acquire();							// if DIERR_INPUTLOST, reacquire and try again
	}
	if (nRet < 0) {
		return 1;
	}

	// This joystick had been successfully Read this frame:
	bJoystickRead[i] = 1;
	
	return 0;
}

// Read the keyboard
static inline int ReadKeyboard()
{
	int nRet = 0;
	if (Keyboard == NULL) {
		return 1;
	}
	if (bKeyboardRead) {							// already read this frame - ready to go
		return 0;
	}

	// Read keyboard, reacquiring if neccesary
	for (int j = 0; j < 2; j++) {
		nRet = Keyboard->GetDeviceState(sizeof(KeyboardState), (void*)KeyboardState);
		if (nRet != DIERR_INPUTLOST) {
			break;
		}
		Keyboard->Acquire();						// if DIERR_INPUTLOST, reacquire and try again
	}
	if (nRet < 0) {									// Failed to read the keyboard
		return 1;
	}

	// The keyboard has been successfully Read this frame:
	bKeyboardRead = 1;

	return 0;
}

// Check a subcode (the 40xx bit in 4001, 4102 etc) for a joystick input code
static int JoystickState(DIJOYSTATE *pState, unsigned int nSubCode)
{
	if (nSubCode < 0x10) {									// Joystick directions
		const int DEADZONE = 0x8000;
		
		switch (nSubCode) {
			case 0x00: return pState->lX < -DEADZONE;		// Left
			case 0x01: return pState->lX > DEADZONE;		// Right
			case 0x02: return pState->lY < -DEADZONE;		// Up
			case 0x03: return pState->lY > DEADZONE;		// Down
			case 0x04: return pState->lZ < -DEADZONE;
			case 0x05: return pState->lZ > DEADZONE;
			case 0x06: return pState->lRx < -DEADZONE;
			case 0x07: return pState->lRx > DEADZONE;
			case 0x08: return pState->lRy < -DEADZONE;
			case 0x09: return pState->lRy > DEADZONE;
			case 0x0A: return pState->lRz < -DEADZONE;
			case 0x0B: return pState->lRz > DEADZONE;
			case 0x0C: return pState->rglSlider[0] < -DEADZONE;
			case 0x0D: return pState->rglSlider[0] > DEADZONE;
			case 0x0E: return pState->rglSlider[1] < -DEADZONE;
			case 0x0F: return pState->rglSlider[1] > DEADZONE;
		}
	}
	if (nSubCode < 0x20) {									// POV hat controls
		DWORD nHatDirection = pState->rgdwPOV[(nSubCode & 0x0F) >> 2];
		if (nHatDirection != (DWORD)-1) {
			switch (nSubCode & 3) {
				case 0:										// Left
					return nHatDirection >= 22500 && nHatDirection <= 31500;
				case 1:										// Right
					return nHatDirection >= 4500 && nHatDirection <= 13500;
				case 2:										// Up
					return nHatDirection >= 31500 || nHatDirection <= 4500;
				case 3:										// Down
					return nHatDirection >= 13500 && nHatDirection <= 22500;
			}
		}
		return 0;
	}
	if (nSubCode < 0x80) {									// Undefined
		return 0;
	}
	if (nSubCode < 0x80 + DINPUT_BUTTON_COUNT) {			// Joystick buttons
		return (pState->rgbButtons[nSubCode & 0x7F] & 0x80) ? 1 : 0;
	}
	
	return 0;
}

// Get the state (pressed = 1, not pressed = 0) of a particular input code
int DinpState(int nCode)
{
	if (nCode < 0) {
		return 0;
	}
	if (nCode < 0x100) {
		if (ReadKeyboard()!= 0) {							// Check keyboard has been read - return not pressed on error
			return 0;
		}
		return(KeyboardState[nCode] & 0x80) ? 1 : 0;		// Return key state
	}

	if (nCode < 0x4000) {
		return 0;
	}
	if (nCode < 0x8000) {
		// Codes 4000-8000 = Joysticks
		int nJoyNumber = 0;
		nJoyNumber = (nCode - 0x4000) >> 8;
		if (nJoyNumber >= nJoystickCount) {					// This joystick number isn't connected
			return 0;
		}
		if (ReadJoystick(nJoyNumber) != 0) {				// There was an error polling the joystick
			return 0;
		}
		
		// Find the joystick state in our array
		return JoystickState(JoyState + nJoyNumber, nCode & 0xFF);
	}

	return 0;
}

// This function finds which key is pressed, and returns its code
int DinpFind(bool CreateBaseline)
{
	int nRetVal = -1;										// assume nothing pressed
	DinpStart();											// reset the flags which say each joystick was Read

	// check if any keyboard keys are pressed
	if (ReadKeyboard() == 0) {
		for (int i = 0; i < 0x100; i++) {
			if (KeyboardState[i] & 0x80) {
				nRetVal = i;
				goto End;
			}
		}
	}

	// Now check all the connected joysticks
	for (int i = 0; i < nJoystickCount; i++) {
		unsigned int j;
		DIJOYSTATE* pState = NULL;
		if (ReadJoystick(i) != 0) {							// There was an error polling the joystick
			continue;
		}
		pState = JoyState + i;
		for (j = 0; j < 0x0F; j++) {						// Axes
			int nDelta = JoyPrevAxes[(i << 3) + (j >> 1)] - DinpAxis(i, (j >> 1));
			if (nDelta < -0x4000 || nDelta > 0x4000) {
				if (JoystickState(pState, j)) {
					nRetVal = 0x4000 | (i << 8) | j;
					goto End;
				}
			}
		}
		
		for (j = 0x10; j < 0x1F; j++) {						// POV hats
			if (JoystickState(pState, j)) {
				nRetVal = 0x4000 | (i << 8) | j;
				goto End;
			}
		}
		
		for (j = 0x80; j < 0x80 + DINPUT_BUTTON_COUNT; j++) {
			if (JoystickState(pState, j)) {
				nRetVal = 0x4000 | (i << 8) | j;
				goto End;
			}
		}
	}

End:

	if (CreateBaseline) {
		for (int i = 0; i < nJoystickCount; i++) {
			for (int j = 0; j < 8; j++) {
				JoyPrevAxes[(i << 3) + j] = DinpAxis(i, j);
			}
		}
	}

	return nRetVal;
}

// Read the axis of a joystick
int DinpAxis(int i, int nAxis)
{
	if (i < 0 || i >= nJoystickCount) {				// This joystick number isn't connected
		return 0;
	}
	if (ReadJoystick(i) != 0) {						// There was an error polling the joystick
		return 0;
	}
	
	switch (nAxis) {
		case 0:
			return JoyState[i].lX;
		case 1:
			return JoyState[i].lY;
		case 2:
			return JoyState[i].lZ;
		case 3:
			return JoyState[i].lRx;
		case 4:
			return JoyState[i].lRy;
		case 5:
			return JoyState[i].lRz;
		case 6:
			return JoyState[i].rglSlider[0];
		case 7:
			return JoyState[i].rglSlider[1];
	}
	
	return 0;
}
