// Burner Input module
#include "app.h"

// 1 if current input is on (i.e. emu in foreground)
static int bCinpOkay = 0;

int CinpState(int nCode)
{
	// Return off for keyboard inputs if current input is turned off
	if ((nCode & 0x4000) == 0 && bCinpOkay == 0) {
		return 0;
	}
	
	// Read from Direct Input
	return DinpState(nCode);
}

// Read the axis of a joystick
int CinpAxis(int i, int nAxis)
{
	// Return off, if current input is turned off
	if (bCinpOkay == 0) {
		return 0;
	}
	
	// Read from Direct Input
	return DinpAxis(i, nAxis);
}

int InputMake(bool bCopy)
{
	bCinpOkay = 0;

	DinpStart();								// Poll joysticks etc

	if (AppProcessKeyboardInput()) {
		bCinpOkay = 1;
	}

	struct GameInp* pgi;
	unsigned int i;

	for (i = 0, pgi = GameInp; i < nGameInpCount; i++, pgi++) {
		if (pgi->Input.pVal == NULL) {
			continue;
		}

		switch (pgi->nInput) {
			case 0:									// Undefined
				pgi->Input.nVal = 0;
				break;
			case 1:									// Constant value
				pgi->Input.nVal = pgi->Input.Constant.nConst;
				break;
			case 2:	{								// Digital input
				int s = CinpState(pgi->Input.Switch.nCode);
				if (pgi->nType == 1) {
					// Set analog controls to full
					if (s) {
						pgi->Input.nVal = 0xFF;
					} else {
						pgi->Input.nVal = 0x01;
					}
				} else {
					// Binary controls
					if (s) {
						pgi->Input.nVal = 1;
					} else {
						pgi->Input.nVal = 0;
					}
				}

				break;
			}
			case 3:	{								// Joystick axis
				int nJoy = CinpAxis(pgi->Input.JoyAxis.nJoy, pgi->Input.JoyAxis.nAxis);
				nJoy >>= 9;
				nJoy += 0x80;

				// Clip axis to 8 bits
				if (nJoy < 0x01) {
					nJoy = 0x01;
				}
				if (nJoy > 0xFF) {
					nJoy = 0xFF;
				}
				pgi->Input.nVal = (unsigned char)nJoy;

				break;
			}
			case 4:	{								// Joystick axis Lo
				int nJoy = CinpAxis(pgi->Input.JoyAxis.nJoy, pgi->Input.JoyAxis.nAxis);
				if (nJoy < 127) {
					nJoy = -nJoy;
					nJoy >>= 8;

					if (nJoy < 0x00) {
						nJoy = 0x00;
					}
					if (nJoy > 0xFF) {
						nJoy = 0xFF;
					}

					pgi->Input.nVal = (unsigned char)nJoy;
				} else {
					pgi->Input.nVal = 0x00;
				}
				
				break;
			}
			case 5:	{								// Joystick axis Hi
				int nJoy = CinpAxis(pgi->Input.JoyAxis.nJoy, pgi->Input.JoyAxis.nAxis);
				if (nJoy > 127) {
					nJoy >>= 8;

					if (nJoy < 0x00) {
						nJoy = 0x00;
					}
					if (nJoy > 0xFF) {
						nJoy = 0xFF;
					}

					pgi->Input.nVal = (unsigned char)nJoy;
				} else {
					pgi->Input.nVal = 0x00;
				}
				
				break;
			}
			case 6:									// Keyboard slider
			case 7:									// Joystick slider
				pgi->Input.nVal = (unsigned char)(pgi->Input.Slider.nSliderValue >> 8);
				break;
		}
		
		if (bCopy) {
			*(pgi->Input.pVal) = pgi->Input.nVal;
		}
	}

	for (i = 0; i < nMacroCount; i++, pgi++) {
		if (pgi->Macro.nMode) {						// Macro is defined
		
			if (bCopy && CinpState(pgi->Macro.Switch.nCode)) {
				for (int j = 0; j < 4; j++) {
					if (pgi->Macro.pVal[j]) {
						*(pgi->Macro.pVal[j]) = pgi->Macro.nVal[j];
					}
				}
			}
		}
	}

	return 0;
}

// Do one frames worth of keyboard input sliders
int InputTick()
{
	unsigned int i;
	struct GameInp *pgi;

	for (i = 0, pgi = GameInp; i < nGameInpCount; i++, pgi++) {
		int nAdd = 0;
		if (pgi->nInput != 6 && pgi->nInput != 7) {					// not a slider
			continue;
		}

		if (pgi->nInput == 6) {
			// Get states of the two keys
			if (CinpState(pgi->Input.Slider.SliderAxis.nSlider[0]))	{
				nAdd -= 0x100;
			}
			if (CinpState(pgi->Input.Slider.SliderAxis.nSlider[1]))	{
				nAdd += 0x100;
			}
		}

		if (pgi->nInput == 7) {
			// Get state of the axis
			nAdd = CinpAxis(pgi->Input.Slider.JoyAxis.nJoy, pgi->Input.Slider.JoyAxis.nAxis);
			nAdd /= 0x100;
		}
		
		// nAdd is now -0x100 to +0x100

		// Change to slider speed
		nAdd *= pgi->Input.Slider.nSliderSpeed;
		nAdd /= 0x100;

		if (pgi->Input.Slider.nSliderCenter) {						// Attact to center
			int v = pgi->Input.Slider.nSliderValue - 0x8000;
			v *= (pgi->Input.Slider.nSliderCenter - 1);
			v /= pgi->Input.Slider.nSliderCenter;
			v += 0x8000;
			pgi->Input.Slider.nSliderValue = v;
		}

		pgi->Input.Slider.nSliderValue += nAdd;
		// Limit slider
		if (pgi->Input.Slider.nSliderValue < 0x0100) {
			pgi->Input.Slider.nSliderValue = 0x0100;
		}
		if (pgi->Input.Slider.nSliderValue > 0xFF00) {
			pgi->Input.Slider.nSliderValue = 0xFF00;
		}
	}
	return 0;
}

int InputInit()
{
	return 0;
}

int InputExit()
{
	return 0;
}
