// Functions for recording & replaying input
#include "app.h"
#include "dynhuff.h"
#include <commdlg.h>

int nReplayStatus = 0;

static unsigned int nStartFrame;
static unsigned int nEndFrame;

static FILE* fp = NULL;
static int nSizeOffset;

static char nPrevInputs[0x0100];

int RecordInput()
{
	struct BurnInputInfo bii;
	memset(&bii, 0, sizeof(bii));
	
	for (unsigned int i = 0; i < nGameInpCount; i++) {
		BurnDrvGetInputInfo(&bii, i);
		if (bii.pVal) {
			if (*bii.pVal != nPrevInputs[i]) {
				EncodeBuffer(i);
				EncodeBuffer(*bii.pVal);
				nPrevInputs[i] = *bii.pVal;
			}
		}
	}
	
	EncodeBuffer(0xFF);
	return 0;
}

int ReplayInput()
{
	unsigned char i;
	struct BurnInputInfo bii;
	memset(&bii, 0, sizeof(bii));
	
	while ((i = DecodeBuffer()) != 0xFF) {
		BurnDrvGetInputInfo(&bii, i);
		if (bii.pVal) {
			*bii.pVal = DecodeBuffer();
		}
else {
			DecodeBuffer();
		}
	}
	if (end_of_buffer) {
		StopReplay();
		return 1;
	} else {
		return 0;
	}
}

static void MakeOfn()
{
	memset(&ofn, 0, sizeof(ofn));
	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = hScrnWnd;
	ofn.lpstrFilter = "Final Burn Input Recording Files (*.fr)\0*.fr\0\0";
	ofn.lpstrFile = szChoice;
	ofn.nMaxFile = sizeof(szChoice);
	ofn.lpstrInitialDir = ".\\replay";
	ofn.Flags = OFN_NOCHANGEDIR | OFN_HIDEREADONLY;
	ofn.lpstrDefExt = "fr";

	return;
}

int StartRecord()
{
	int nRet;
	int bOldPause;

	sprintf(szChoice, "%.8s", BurnDrvText(0));
	MakeOfn();
	ofn.lpstrTitle = "Record Input to File";
	ofn.Flags |= OFN_OVERWRITEPROMPT;

	bOldPause = bRunPause;
	bRunPause = 1;
	nRet = GetSaveFileName(&ofn);
	bRunPause = bOldPause;

	if (nRet == 0) {
		return 1;
	}
	
	{
		const char szFileHeader[] = "FB1 ";				// File identifier
		fp = fopen(szChoice, "wb");
		
		nRet = 0;
		if (fp == NULL) {
			nRet = 1;
		} else {
			fwrite(&szFileHeader, 1, 4, fp);
			nRet = StateSaveEmbed(fp, -1, 1);
			if (nRet >= 0) {
				const char szChunkHeader[] = "FR1 ";	// Chunk identifier
				int nZero = 0;
				
				fwrite(&szChunkHeader, 1, 4, fp);		// Write chunk identifier
				
				nSizeOffset = ftell(fp);
				
				fwrite(&nZero, 1, 4, fp);				// reserve space for chunk size

				fwrite(&nZero, 1, 4, fp);				// reserve space for number of frames
				
				fwrite(&nZero, 1, 4, fp);				// reserved
				fwrite(&nZero, 1, 4, fp);				//
				fwrite(&nZero, 1, 4, fp);				//

				nRet = EmbedCompressedFile(fp, -1);
			}
		}
	}
	
	if (nRet) {
		fclose(fp);
		fp = NULL;
		
		AppError("Error Creating File", 1);
		return 1;
	} else {
		struct BurnInputInfo bii;
		memset(&bii, 0, sizeof(bii));
		
		nReplayStatus = 1;								// Set record status
		MenuEnableItems();

		nStartFrame = GetCurrentFrame();
		
		// Create a baseline so we can just record the deltas afterwards
		for (unsigned int i = 0; i < nGameInpCount; i++) {
			BurnDrvGetInputInfo(&bii, i);
			if (bii.pVal) {
				EncodeBuffer(*bii.pVal);
				nPrevInputs[i] = *bii.pVal;
			} else {
				EncodeBuffer(0);
			}
		}				

#ifdef _DEBUG
		printf("*** Recording of file %s started.\n", szChoice);
#endif
		return 0;
	}
}

int StartReplay(const char* szFileName)					// const char* szFileName = NULL
{
	int nRet;
	int bOldPause;
	
	if (szFileName) {
		strcpy(szChoice, szFileName);
	} else {
		if (bDrvOkay) {
			sprintf(szChoice, "%.8s*.fr", BurnDrvText(0));
		} else {
			sprintf(szChoice, "*.fr");
		}
		MakeOfn();
		ofn.lpstrTitle = "Replay Input from File";

		bOldPause = bRunPause;
		bRunPause = 1;
		nRet = GetOpenFileName(&ofn);
		bRunPause = bOldPause;

		if (nRet == 0) {
			return 1;
		}
	}
	
	{
		const char szFileHeader[] = "FB1 ";					// File identifier
		char ReadHeader[] = "    ";
		fp = fopen(szChoice, "rb");
		memset(ReadHeader, 0, 4);
		fread(ReadHeader, 1, 4, fp);						// Read identifier
		if (memcmp(ReadHeader, szFileHeader, 4)) {			// Not the right file type
			fclose(fp);
			fp = NULL;
			nRet = 2;
		} else {
			// First load the savestate associated with the recording		
			nRet = StateLoadEmbed(fp, -1, 1);
			if (nRet == 0) {
				const char szChunkHeader[] = "FR1 ";		// Chunk identifier
				fread(ReadHeader, 1, 4, fp);				// Read identifier
				if (memcmp(ReadHeader, szChunkHeader, 4)) {	// Not the right file type
					fclose(fp);
					fp = NULL;
					nRet = 2;
				} else {
					// Open the recording itself
					fseek(fp, 0x04, SEEK_CUR);				// Skip chunk size
					fread(&nEndFrame, 1, 4, fp);			// Read framecount
					nStartFrame = GetCurrentFrame();
					nEndFrame += nStartFrame;
					fseek(fp, 0x0C, SEEK_CUR);				// Skip rest of header
					nRet = EmbedCompressedFile(fp, -1);			
					
				}
			}
		}
		
		// Describe any possible errors:		
		if (nRet == 3) {
			AppError("This input recording is for an unavailable game.", 0);
		} else {
			if (nRet == 4) {
				AppError("This input recording is too old and cannot be loaded.", 0);
			} else {
				if (nRet == 5) {
					AppError("Emulator is too old to load this input recording.", 0);
				} else {
					if (nRet) {
						AppError("Error loading input recording.", 0);
					}
				}
			}
		}
		
		if (nRet) {
			fclose(fp);
			fp = NULL;
			
			return 1;
		}
	}
		
	nReplayStatus = 2;							// Set replay status
	MenuEnableItems();
	
	{
		struct BurnInputInfo bii;
		memset(&bii, 0, sizeof(bii));
		
		LoadCompressedFile();
		
		// Get the baseline
		for (unsigned int i = 0; i < nGameInpCount; i++) {
			BurnDrvGetInputInfo(&bii, i);
			if (bii.pVal) {
				*bii.pVal = DecodeBuffer();
			} else {
				DecodeBuffer();
			}
		}				
	}

#ifdef _DEBUG
	printf("*** Replay of file %s started.\n", szChoice);
#endif
	return 0;
}

static void CloseRecord()
{
	int nFrames = GetCurrentFrame() - nStartFrame;

	WriteCompressedFile();
	
	fseek(fp, 0, SEEK_END);	
	int nChunkSize = ftell(fp) - 4 - nSizeOffset;		// Fill in chunk size and no of recorded frames
	fseek(fp, nSizeOffset, SEEK_SET);
	fwrite(&nChunkSize, 1, 4, fp);
	fwrite(&nFrames, 1, 4, fp);
	
	fclose(fp);
	fp = NULL;
}

static void CloseReplay()
{
	CloseCompressedFile();
	
	fclose(fp);
	fp = NULL;
}

void StopReplay()
{
	if (nReplayStatus) {		
		if (nReplayStatus == 1) {
#ifdef _DEBUG
			printf(" ** Recording stopped, recorded %d frames.\n", GetCurrentFrame() - nStartFrame);
#endif
			CloseRecord();
#ifdef _DEBUG
			PrintResult();
#endif
		} else {
#ifdef _DEBUG
			printf(" ** Replay stopped, replayed %d frames.\n", GetCurrentFrame() - nStartFrame);
#endif
			CloseReplay();
		}
		nReplayStatus = 0;
		MenuEnableItems();
	}
}

