// DirectDraw Three Output (dto) code
#include "app.h"
#include "directx_support.h"

static IDirectDraw7 *DtoDD = NULL;				// DirectDraw interface
static IDirectDrawSurface7 *DtoPrim = NULL;		// Primary surface
static IDirectDrawSurface7 *DtoBack = NULL;		// Back buffer surface

static bool bRotateGame;
static bool bRotateScanlines;

static int nDtosWidth = 0, nDtosHeight = 0;		// bitmap size
static int nGameWidth = 0, nGameHeight = 0;		// screen size
static int nDtosDepth = 0, nDtosBpp = 0;
static int bDtosScan = 0;
static unsigned char* DtosMem = NULL;			// Memory buffer
static int nDtosMemPitch = 0;					// Memory buffer pitch
static IDirectDrawSurface7* pddsDtos = NULL;		// The screen surface
static int nHalfMask = 0;

static int nUseSys;								// Use System or Video memory

static int DtosInit();
static int DtosExit();

static int DtoClear(IDirectDrawSurface7 *Surf)
{
	// Clear a surface to black
	DDBLTFX BltFx;

	if (Surf == NULL) {
		return 1;
	}

	memset(&BltFx, 0, sizeof(BltFx));
	BltFx.dwSize = sizeof(BltFx);

	if (Surf->Blt(NULL,NULL,NULL,DDBLT_COLORFILL,&BltFx) < 1) {
		return 1;
	} else {
		return 0;
	}
}

static int DtoPrimClear()
{
	if (DtoPrim == NULL) {
		return 1;
	}
	DtoClear(DtoPrim);					// Clear 1st page

	if (DtoBack == NULL) {
		return 0;
	}
	DtoClear(DtoBack);					// Clear 2nd page
	DtoPrim->Flip(NULL, DDFLIP_WAIT);	// wait till the flip actually occurs
	DtoClear(DtoBack);					// Clear 3rd page

	return 0;
}

static int DtoPrimInit(int bTriple)
{
	DDSURFACEDESC2 ddsd;
	int nRet = 0;

	// Create the primary surface
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	if (bTriple) {
		// Make a primary surface capable of triple bufferring
		ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX | DDSCAPS_VIDEOMEMORY;
		ddsd.dwBackBufferCount = 2;
	} else {
		ddsd.dwFlags = DDSD_CAPS;
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	}

	DtoPrim = NULL;
	nRet = DtoDD->CreateSurface(&ddsd, &DtoPrim, NULL);
	if (nRet < 0 || DtoPrim == NULL) {
		return 1;
	}

	if (bTriple == 0) {						// We are done
		return 0;
	}

	// Get the back buffer
	memset(&ddsd.ddsCaps, 0, sizeof(ddsd.ddsCaps));
	ddsd.ddsCaps.dwCaps = DDSCAPS_BACKBUFFER;

	DtoBack = NULL;
	nRet = DtoPrim->GetAttachedSurface(&ddsd.ddsCaps, &DtoBack);
	if (nRet < 0 || DtoBack == NULL) {		// Failed to make triple buffer
		RELEASE(DtoPrim)
		return 1;
	}

	DtoPrimClear();							// Clear all three pages
	return 0;
}

static int DtoClipInit()
{
	IDirectDrawClipper *pddcClipper = NULL;			// The clipper
	
	DirectDrawCreateClipper(0, &pddcClipper, NULL);

	if (pddcClipper) {
		if (pddcClipper->SetHWnd(0, hVidWnd) >= 0) {
			DtoPrim->SetClipper(pddcClipper);
			RELEASE(pddcClipper);
		}
	}
	return 0;
}

static int DtoExit();

static int DtoInit()
{
	hVidWnd = hScrnWnd;						// Use Screen window for video

	// Get pointer to DirectDraw device
	DirectDrawCreateEx(NULL, (void**)&DtoDD, IID_IDirectDraw7, NULL);

	if (BurnDrvGetFlags() & BDF_ROTATE_GRAPHICS_CCW) {
		bRotateGame = true;
	} else {
		bRotateGame = false;
	}
		
	DtoPrim = NULL;							// No primary surface yet
	DtoBack = NULL;

	// Remember the changes to the display
	if (nVidFullscreen) {
		if (EnterFullscreenMode(DtoDD, nScreenSize, 0, NULL, NULL)) {
			DtoExit();
			return 1;
		}

		if (bVidTripleBuffer) {
			if (DtoPrimInit(1)) {			// Try to make triple buffer
				AppError("Could not make Triple Buffer", 1);
				// If we fail, fail entirely and make a normal buffer
				RELEASE(DtoBack)
				RELEASE(DtoPrim)
			}
		}
	} else {
		DtoDD->SetCooperativeLevel(hVidWnd, DDSCL_NORMAL);
	}

	if (DtoPrim == NULL) {
		// No primary surface yet, so try normal
		if (DtoPrimInit(0)) {
			DtoExit();
			return 1;
		}
	}

	// Make a clipper in windowed mode
	if (nVidFullscreen == 0) {
		DtoClipInit();
	}

	VidSSetupGamma(DtoPrim);
	
	// Init the buffer surfaces
	DtosInit();

	return 0;
}

static int DtoExit()
{
	VidSRestoreGamma();

	RestoreScreenMode();

	DtosExit();

	RELEASE(DtoPrim)					// a single call releases all surfaces
	DtoBack = NULL;

	RELEASE(DtoDD)
	
	return 0;
}

// Create a buffer in normal memory
static int DtosMakeMem()
{
	int nMemLen = 0;
	
	nDtosMemPitch = nDtosWidth * nDtosBpp;
	nMemLen = nDtosHeight * nDtosMemPitch;
	DtosMem = (unsigned char*)malloc(nMemLen);

	if (DtosMem) {
		memset(DtosMem, 0, nMemLen);
		return 0;
	} else {
		return 1;
	}
}

static int AutodetectUseSys()
{
	// Try to autodetect the best secondary buffer type to use, based on the cards capabilities
	DDCAPS ddc;
	if (DtoDD == NULL) {
		return 1;
	}
	memset(&ddc, 0, sizeof(ddc));
	ddc.dwSize = sizeof(ddc);
	DtoDD->GetCaps(&ddc, NULL);

	if (ddc.dwCaps & DDCAPS_BLTSTRETCH) {	// If it can do a hardware stretch use video memory
		return 0;
	} else {								// Otherwise use system memory:
		return 1;
	}
}

// Create a secondary DD surface for the screen
static int DtosMakeSurf()
{
	int nRet;
	nUseSys = 0;
	DDSURFACEDESC2 ddsd;

	if (DtoDD == NULL) {
		return 1;
	}

	nUseSys = nUseSysMemory;
	if (nUseSys < 0) {
		nUseSys = AutodetectUseSys();
	}

	// Try to allocate buffer in Video memory first, if that fails use System memory
	for (nRet = -1; (nUseSys <= 1) && (nRet < 0); nUseSys++) {
		memset(&ddsd, 0, sizeof(ddsd));
		ddsd.dwSize = sizeof(ddsd);
		ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
		ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
		if (nUseSys == 0) {
			ddsd.ddsCaps.dwCaps |= DDSCAPS_VIDEOMEMORY;
		} else {
			if (nUseSys == 1) {
				ddsd.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;
			}
		}

		ddsd.dwWidth = nGameWidth;
		ddsd.dwHeight = nGameHeight;
		if (bDtosScan) {
			if (bRotateScanlines) {
				ddsd.dwWidth <<= 1;
			} else {
				ddsd.dwHeight <<= 1;
			}
		}

		nRet = DtoDD->CreateSurface(&ddsd, &pddsDtos, NULL);
	}
	nUseSys -= 1;

	if (nRet < 0 || pddsDtos == NULL) {
		return 1;
	}
	
	DtoClear(pddsDtos);

	return 0;
}

static int DtosInit()
{
	if (DtoDD == NULL || bDrvOkay == 0) {
		return 1;
	}

	// Get the game screen size
	BurnDrvGetScreen(&nGameWidth, &nGameHeight);
	
	if (bRotateGame) {
		nDtosWidth = nGameHeight;
		nDtosHeight = nGameWidth;
	} else {
		nDtosWidth = nGameWidth;
		nDtosHeight = nGameHeight;
	}

	if (bVidScanRotate && nGameWidth < nGameHeight) {
		bRotateScanlines = true;
	} else {
		bRotateScanlines = false;
	}

	RECT rect = { 0, 0, 0, 0 };
	GetClientScreenRect(hVidWnd, &rect);
	if (!nVidFullscreen) {
		rect.top += nMenuHeight;
	}
	if (!bVidFullStretch) {
		VidSScaleImage(&rect, nGameWidth, nGameHeight, bVidScanRotate);
	}
	
	bDtosScan = bVidScanlines;
	if (bRotateScanlines) {
		if (rect.right - rect.left < 2 * nGameWidth) {
			bDtosScan = 0;
		}
	} else {
		if (rect.bottom - rect.top < 2 * nGameHeight) {
			bDtosScan = 0;
		}
	}
	
	nDtosDepth = VidSGetSurfaceDepth(DtoPrim);	// Get color depth of primary surface
	nDtosBpp = (nDtosDepth + 7) >> 3;
	nBurnBitDepth = nDtosDepth;
	nBurnBpp = nDtosBpp;						// Set Burn library Bytes per pixel

	// Make the mask to mask out all but the lowest intensity bit
	nHalfMask = 0;
	if (nDtosDepth == 15) {
		nHalfMask = 0xfbdefbde;
	} else {
		if (nDtosDepth == 16) {
			nHalfMask = 0xf7def7de;
		} else {
			nHalfMask = 0xfefefefe;
		}
	}

	// Use our callback to get colors:
	SetBurnHighCol(nDtosDepth);
	// Recalculate the whole palette next time BurnDraw is called
	// This will use our callback for the whole palette
	BurnRecalcPal();

	// Make the normal memory buffer
	if (DtosMakeMem()) {
		DtosExit();
		return 1;
	}

	// Make the DirectDraw secondary surface
	if (DtosMakeSurf()) {
		DtosExit();
		return 1;
	}

	if (DtoBack) {
		int nFlags = 0;
		if (rect.right - rect.left < 1280) {
			nFlags++;
			if (rect.right - rect.left < 800) {
				nFlags++;
				if (rect.right - rect.left < 640) {
					nFlags++;
				}
			}
		}
		VidSInitOSD(DtoDD, nFlags);
	} else {
		int nFlags = 3;
		if (bDtosScan) {
			if (bVidScanRotate) {
				nFlags |= 0x10;
			} else {
				nFlags |= 0x20;
			}
		}
		VidSInitOSD(DtoDD, nFlags);
	}

	return 0;
}

static int DtosExit()
{
	RELEASE(pddsDtos)
	nDtosMemPitch = 0;
	if (DtosMem) {
		free(DtosMem);
		DtosMem = NULL;
	}
	nHalfMask = 0; nDtosBpp = 0; nDtosDepth = 0;
	nDtosWidth = 0; nDtosHeight = 0;
	bDtosScan = 0;

	// Destroy resources allocated for "OSD"
	VidSExitOSD();

	return 0;
}

// Copy DtosMem to pddsDtos, rotate
static int DtosMemToSurfRotate(DDSURFACEDESC2* ddsd)
{
	unsigned char *pd, *ps, *pdd;
	unsigned char *Surf;
	int nPitch;

	Surf = (unsigned char*)ddsd->lpSurface;
	nPitch = ddsd->lPitch;

	pd = Surf;
	
	if (bDtosScan) {
		if (bVidScanHalf) {
			if (bVidScanRotate) {				
				switch (nDtosBpp) {
					case 4: {
						int t;
						for (int y = 0; y < nGameHeight; y++, pd += nPitch) {
							ps = DtosMem + (nGameHeight - 1 - y) * 4;
							pdd = pd;
							for (int x = 0; x < nGameWidth; x++) {
								t = *(int*)ps;
								ps += nDtosMemPitch;
								*(int*)pdd = t;
								pdd += 4;
								*(int*)pdd = (t & 0xFEFEFE) >> 1;
								pdd += 4;
							}
						}
						break;
					}
					case 3: {
						for (int y = 0; y < nGameHeight; y++, pd += nPitch) {
							ps = DtosMem + (nGameHeight - 1 - y) * 3;
							pdd = pd;
							for (int x = 0; x < nGameWidth; x++) {
								pdd[0] = ps[0];
								pdd[1] = ps[1];
								pdd[2] = ps[2];
								pdd[3] = ps[0] >> 1;
								pdd[4] = ps[1] >> 1;
								pdd[5] = ps[2] >> 1;
								ps += nDtosMemPitch;
								pdd +=6;
							}
						}
						break;
					}
					case 2: {
						short t;
						for (int y = 0; y < nGameHeight; y++, pd += nPitch) {
							ps = DtosMem + (nGameHeight - 1 - y) * 2;
							pdd = pd;
							for (int x = 0; x < nGameWidth; x++) {
								t = *(short*)ps;
								ps += nDtosMemPitch;
								*(short*)pdd = t;
								pdd += 2;
								*(short*)pdd = (t & nHalfMask) >> 1;
								pdd += 2;
							}
						}
						break;
					}
				}

			} else {
				unsigned char* pBuffer = (unsigned char*)malloc(nPitch);
				unsigned char* pd2;
				switch (nDtosBpp) {
					case 4: {
						int t;
						for (int y = 0; y < nGameHeight; y++, pd += nPitch * 2) {
							ps = DtosMem + (nGameHeight - 1 - y) * 4;
							pdd = pd;
							pd2 = pBuffer;
							for (int x = 0; x < nGameWidth; x++) {
								t = *(int*)ps;
								ps += nDtosMemPitch;
								*(int*)pdd = t;
								*(int*)pd2 = t;
								pdd += 4;
								pd2 += 4;
							}
							pdd = pd + nPitch;
							for (ps = pBuffer; ps < (pBuffer + nGameWidth * 4); ps += 4) {
								*(int*)pdd = (*(int*)ps & 0xFEFEFE) >> 1;
								pdd += 4;
							}
						}
						break;
					}
					case 3: {
						for (int y = 0; y < nGameHeight; y++, pd += nPitch * 2) {
							ps = DtosMem + (nGameHeight - 1 - y) * 3;
							pdd = pd;
							pd2 = pBuffer;
							for (int x = 0; x < nGameWidth; x++) {
								pdd[0] = ps[0];
								pd2[0] = ps[0];
								pdd[1] = ps[1];
								pd2[1] = ps[1];
								pdd[2] = ps[2];
								pd2[2] = ps[2];
								ps += nDtosMemPitch;
								pdd += 3;
								pd2 += 3;
							}
						}
						pdd = pd + nPitch;
						for (ps = pBuffer; ps < (pBuffer + nGameWidth * 3); ps += 3) {
							pdd[0] = ps[0] >> 1;
							pdd[1] = ps[1] >> 1;
							pdd[2] = ps[2] >> 1;
							pdd += 3;
						}
						break;
					}
					case 2: {
						short t;
						for (int y = 0; y < nGameHeight; y++, pd += nPitch * 2) {
							ps = DtosMem + (nGameHeight - 1 - y) * 2;
							pdd = pd;
							pd2 = pBuffer;
							for (int x = 0; x < nGameWidth; x++) {
								t = *(short*)ps;
								ps += nDtosMemPitch;
								*(short*)pdd = t;
								*(short*)pd2 = t;
								pdd += 2;
								pd2 += 2;
							}
						}
						pdd = pd + nPitch;
						for (ps = pBuffer; ps < (pBuffer + nGameWidth * 2); ps += 2) {
							*(short*)pdd = (*(short*)ps & nHalfMask) >> 1;
							pdd += 2;
						}
						break;
					}
				}
				free (pBuffer);
			}
		} else {
			int nPixelSize = nDtosBpp;
			if (bVidScanRotate) {
				nPixelSize <<= 1;
			} else {
				nPitch <<= 1;
			}
			switch (nDtosBpp) {
				case 4: {
					for (int y = 0; y < nGameHeight; y++, pd += nPitch) {
						ps = DtosMem + (nGameHeight - 1 - y) * 4;
						pdd = pd;
						for (int x = 0; x < nGameWidth; x++) {
							*(int*)pdd = *(int*)ps;
							ps += nDtosMemPitch;
							pdd += nPixelSize;
						}
					}
					break;
				}
				case 3: {
					for (int y = 0; y < nGameHeight; y++, pd += nPitch) {
						ps = DtosMem + (nGameHeight - 1 - y) * 3;
						pdd = pd;
						for (int x = 0; x < nGameWidth; x++) {
							pdd[0] = ps[0];
							pdd[1] = ps[1];
							pdd[2] = ps[2];
							ps += nDtosMemPitch;
							pdd += nPixelSize;
						}
					}
					break;
				}
				case 2: {
					for (int y = 0; y < nGameHeight; y++, pd += nPitch) {
						ps = DtosMem + (nGameHeight - 1 - y) * 2;
						pdd = pd;
						for (int x = 0; x < nGameWidth; x++) {
							*(short*)pdd = *(short*)ps;
							ps += nDtosMemPitch;
							pdd += nPixelSize;
						}
					}
					break;
				}
			}
		}
	} else {
		switch (nDtosBpp) {
			case 4: {
				for (int y = 0; y < nGameHeight; y++, pd += nPitch) {
					ps = DtosMem + (nGameHeight - 1 - y) * 4;
					pdd = pd;
					for (int x = 0; x < nGameWidth; x++) {
						*(int*)pdd = *(int*)ps;
						ps += nDtosMemPitch;
						pdd += 4;
					}
				}
				break;
			}
			case 3: {
				for (int y = 0; y < nGameHeight; y++, pd += nPitch) {
					ps = DtosMem + (nGameHeight - 1 - y) * 3;
					pdd = pd;
					for (int x = 0; x < nGameWidth; x++) {
						pdd[0] = ps[0];
						pdd[1] = ps[1];
						pdd[2] = ps[2];
						ps += nDtosMemPitch;
						pdd +=3;
					}
				}
				break;
			}
			case 2:	{
				for (int y = 0; y < nGameHeight; y++, pd += nPitch) {
					ps = DtosMem + (nGameHeight - 1 - y) * 2;
					pdd = pd;
					for (int x = 0; x < nGameWidth; x++) {
						*(short*)pdd = *(short*)ps;
						ps += nDtosMemPitch;
						pdd += 2;
					}
				}
				break;
			}
		}
	}

	return 0;
}

// Copy DtosMem to pddsDtos, don't rotate, add scanlines in odd lines if needed
static int DtosMemToSurfNoRotate1(DDSURFACEDESC2* ddsd, int nField, int nHalf)
{
	unsigned char *pd, *ps;

	unsigned char* Surf = (unsigned char*)ddsd->lpSurface;
	int nPitch = ddsd->lPitch;
	if (bDtosScan) {
		if (nField) {			// copy to odd fields
			Surf += nPitch;
		}
		nPitch <<= 1;
	}

	pd = Surf; ps = DtosMem;
	for (int y = 0; y < nDtosHeight; y++, pd += nPitch, ps += nDtosMemPitch) {
		if (nHalf == 0) {
			memcpy(pd, ps, nDtosMemPitch);
		} else {
			unsigned char* psEnd = ps + nDtosMemPitch;
			unsigned char* pdp = pd;
			unsigned char* psp = ps;

			do {
				unsigned int t;
				t = *((unsigned int *)(psp));
				t = (t & nHalfMask) >> 1;
				*((unsigned int *)(pdp)) = t;

				psp += 4;
				pdp += 4;
			} while (psp < psEnd);
		}
	}

	return 0;
}

// Copy DtosMem to pddsDtos, don't rotate, add rotated scanlines
static int DtosMemToSurfNoRotate2(DDSURFACEDESC2* ddsd)
{
	unsigned char *pd, *ps;
	unsigned char* Surf = (unsigned char*)ddsd->lpSurface;
	int nPitch = ddsd->lPitch;
	
	if (bVidScanHalf) {
		pd = Surf; ps = DtosMem;
		switch (nDtosBpp) {
			case 4: {
				unsigned char* pdp;
				unsigned char* psp;
				unsigned char* psEnd;
				for (int y = 0; y < nDtosHeight; y++, pd += nPitch, ps += nDtosMemPitch) {
					pdp = pd;
					psp = ps;
					psEnd = ps + nDtosMemPitch;
					do {
						unsigned int t = *(unsigned int*)psp;
						*(unsigned int*)pdp = t;
						psp += 4;
						pdp += 4;
						*(unsigned int*)pdp = (t & 0xFEFEFE) >> 1;
						pdp += 4;
					} while (psp < psEnd);
				}
				break;
			}
			case 3: {
				unsigned char* pdp;
				unsigned char* psp;
				unsigned char* psEnd;
				for (int y = 0; y < nDtosHeight; y++, pd += nPitch, ps += nDtosMemPitch) {
					pdp = pd;
					psp = ps;
					psEnd = ps + nDtosMemPitch;
					do {
						pdp[0] = psp[0];
						pdp[1] = psp[1];
						pdp[2] = psp[2];
						pdp[3] = psp[0] >> 1;
						pdp[4] = psp[1] >> 1;
						pdp[5] = psp[2] >> 1;
						psp += 3;
						pdp += 6;
					} while (psp < psEnd);
				}
				break;
			}
			case 2: {
				unsigned char* pdp;
				unsigned char* psp;
				unsigned char* psEnd;
				unsigned short t;
				for (int y = 0; y < nDtosHeight; y++, pd += nPitch, ps += nDtosMemPitch) {
					pdp = pd;
					psp = ps;
					psEnd = ps + nDtosMemPitch;
					do {
						t = *(unsigned short*)psp;
						*(unsigned short*)pdp = t;
						psp += 2;
						pdp += 2;
						*(unsigned short*)pdp = (t & nHalfMask) >> 1;
						pdp += 2;
					} while (psp < psEnd);
				}
				break;
			}
		}
		
	} else {
		pd = Surf; ps = DtosMem;
		switch (nDtosBpp) {
			case 4: {
				unsigned char* pdp;
				unsigned char* psp;
				unsigned char* psEnd;
				for (int y = 0; y < nDtosHeight; y++, pd += nPitch, ps += nDtosMemPitch) {
					pdp = pd;
					psp = ps;
					psEnd = ps + nDtosMemPitch;
					do {
						*(unsigned int*)pdp = *(unsigned int*)psp;

						psp += 4;
						pdp += 8;
					} while (psp < psEnd);
				}
				break;
			}
			case 3: {
				unsigned char* pdp;
				unsigned char* psp;
				unsigned char* psEnd;
				for (int y = 0; y < nDtosHeight; y++, pd += nPitch, ps += nDtosMemPitch) {
					pdp = pd;
					psp = ps;
					psEnd = ps + nDtosMemPitch;
					do {
						pdp[0] = psp[0];
						pdp[1] = psp[1];
						pdp[2] = psp[2];

						psp += 3;
						pdp += 6;
					} while (psp < psEnd);
				}
				break;
			}
			case 2: {
				unsigned char* pdp;
				unsigned char* psp;
				unsigned char* psEnd;
				for (int y = 0; y < nDtosHeight; y++, pd += nPitch, ps += nDtosMemPitch) {
					pdp = pd;
					psp = ps;
					psEnd = ps + nDtosMemPitch;
					do {
						*(unsigned short*)pdp = *(unsigned short*)psp;

						psp += 2;
						pdp += 4;
					} while (psp < psEnd);
				}
				break;
			}
		}
	}

	return 0;
}

static int DtosMemToSurf()
{
	DDSURFACEDESC2 ddsd;

	if (pddsDtos == NULL) {
		return 1;
	}

	if (DtoPrim->IsLost()) {		// We've lost control of the screen
		return 1;	
	}
	
	// Lock the surface so we can write to it
	memset(&ddsd, 0, sizeof(ddsd));

	ddsd.dwSize = sizeof(ddsd);
	pddsDtos->Lock(NULL, &ddsd, DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT, NULL);
	if (ddsd.lpSurface == NULL) {
		return 1;
	}

	if (bRotateGame) {
		DtosMemToSurfRotate(&ddsd);
	} else {
		if (bDtosScan && bRotateScanlines) {
			DtosMemToSurfNoRotate2(&ddsd);
		} else {
			DtosMemToSurfNoRotate1(&ddsd, 0, 0);
			if (bDtosScan && bVidScanHalf) {
				DtosMemToSurfNoRotate1(&ddsd, 1, 1);
			}
		}
	}

	pddsDtos->Unlock(NULL);
	return 0;
}

// Run one frame and render the screen
int DtosFrame(bool bRedraw)			// bRedraw = 0
{
	if (DtosMem == NULL) {
		return 1;
	}

	if (DtoPrim->IsLost()) {		// We've lost control of the screen
		if (VidSRestoreOSD()) {
			return 1;
		}
		
		if (FAILED(DtoDD->RestoreAllSurfaces())) {
			return 1;
		}
		
		DtoPrimClear();
	}

	pSShotBuffer = NULL;
	pBurnDraw = DtosMem;
	nBurnPitch = nDtosMemPitch;

	if (bDrvOkay) {
		if (bRedraw) {				// Redraw current frame
			if (BurnDrvRedraw()) {
				BurnDrvFrame();		// No redraw function provided, advance one frame
			}
		} else {
			BurnDrvFrame();			// Run one frame and draw the screen
		}
	} else {
		// Blank the screen
		memset(DtosMem, 0, nDtosMemPitch * nDtosHeight);
	}
	pSShotBuffer = pBurnDraw;
	pBurnDraw = NULL;
	nBurnPitch = 0;
	
	DtosMemToSurf();

	return 0;
}

// Paint the Dtos surface onto the primary surface
static int DtosPaint(int bValidate)
{
	RECT Dest = { 0, 0, 0, 0 };

	if (DtoPrim == NULL || pddsDtos == NULL) {
		return 1;
	}

	if (DtoPrim->IsLost()) {		// We've lost control of the screen
		return 1;	
	}
	
	GetClientScreenRect(hVidWnd, &Dest);
	if (!nVidFullscreen) {
		Dest.top += nMenuHeight;
	}

	if (bVidArcaderes && nVidFullscreen) {
		Dest.left = (Dest.right + Dest.left) / 2;
		Dest.left -= nGameWidth / 2;
		Dest.right = Dest.left + nGameWidth;

		Dest.top = (Dest.top + Dest.bottom) / 2;
		Dest.top -= nGameHeight / 2;
		Dest.bottom = Dest.top + nGameHeight;
	} else {
		if (!bVidFullStretch) {
			VidSScaleImage(&Dest, nGameWidth, nGameHeight, bVidScanRotate);
		}
	}

	if (bValidate & 2) {
		DtosMemToSurf();
	}

	if (DtoBack != NULL) {
		// Triple bufferring
		if (DtoBack->Blt(&Dest, pddsDtos, NULL, DDBLT_ASYNC, NULL) < 0) {
			if (DtoBack->Blt(&Dest, pddsDtos, NULL, DDBLT_WAIT, NULL) < 0) {
				return 1;
			}
		}
		VidSDisplayOSD(DtoBack, &Dest);

		DtoPrim->Flip(NULL, DDFLIP_WAIT);
	} else {
		// Normal
		RECT rect = { 0, 0, nGameWidth, nGameHeight };

		// Display OSD text message
		VidSDisplayOSD(pddsDtos, &rect);

		if (DtoPrim->Blt(&Dest, pddsDtos, NULL, DDBLT_ASYNC, NULL) < 0) {
			if (DtoPrim->Blt(&Dest, pddsDtos, NULL, DDBLT_WAIT, NULL) < 0) {
				return 1;
			}
		}
	}

	if (bValidate & 1) {
		// Validate the rectangle we just drew
		POINT c = {0, 0};
		ClientToScreen(hVidWnd, &c);
		Dest.left -= c.x; Dest.right -= c.x;
		Dest.top -= c.y; Dest.bottom -= c.y;
		ValidateRect(hVidWnd, &Dest);
	}

	return 0;
}

static int DtosScale(RECT* pRect, int nWidth, int nHeight)
{
	return VidSScaleImage(pRect, nWidth, nHeight, bVidScanRotate);
}

// The video output plugin:
struct VidOut VidOutDDraw = {DtoInit, DtoExit, DtosFrame, DtosPaint, DtosScale};

